//-----------------------------------------------------------------------------
// Torque X Game Engine
// Copyright  GarageGames.com, Inc.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;

using System.Xml.Serialization;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using GarageGames.Torque.Core;
using GarageGames.Torque.Util;
using GarageGames.Torque.Materials;
using GarageGames.Torque.SceneGraph;
using GarageGames.Torque.GFX;
using GarageGames.Torque.RenderManager;
using GarageGames.Torque.MathUtil;

namespace GarageGames.Torque.T2D
{
    public class T2DStaticSprite : T2DSceneObject, IDisposable
    {
        //======================================================
        #region Constructors

        public T2DStaticSprite()
        {
            // by default, we'll have physics and collision
            CreateWithCollision = true;
            CreateWithPhysics = true;
        }

        #endregion

        //======================================================
        #region Public Properties
        /// <summary>
        /// Specifies the Sprite's Texture
        /// </summary>
        public RenderMaterial Material
        {
            // forward access to quad object
            get { return _quad.Material; }
            set { _quad.Material = value; _quad.RefillVB = true; }
        }

        /// <summary>
        /// Specifies the region of the material to use, as defined by the material's TextureDivider.
        /// </summary>
        public int MaterialRegionIndex
        {
            get { return _materialRegionIndex; }
            set { _materialRegionIndex = value; _quad.RefillVB = true; }
        }

        [XmlIgnore]
        [TorqueCloneIgnore]
        public Vector2[] Points
        {
            set { _quad.Points = value; }
            get { return _quad.Points; }
        }

        public override string AutoName
        {
            get
            {
                string autoname = base.AutoName;
                // add material name
                if (Material != null)
                    autoname += "_" + Material.AutoName;

                return autoname;
            }
        }

        #endregion

        //======================================================
        #region Public Methods

        public override bool OnRegister()
        {
            if (!base.OnRegister())
                return false;

            _quad.IgnoreTextureCoordsUpdate = true;

            return true;
        }

        public override void Render(SceneRenderState srs)
        {
            Profiler.Instance.StartBlock("T2DStaticSprite.Render");

            Vector2 halfSize = Vector2.Multiply(_size, 0.5f);

            Vector3 ScaleVector = new Vector3(halfSize.X, halfSize.Y, 1);

            Matrix objToWorld = Matrix.Identity;

            // scale
            Matrix ScaleMatrix = Matrix.CreateScale(ScaleVector);

            // translate            
            Matrix TranslationMatrix = Matrix.CreateTranslation(new Vector3(Position.X, Position.Y, LayerDepth));

            // rotate            
            Matrix RotationMatrix = Matrix.CreateRotationZ(MathHelper.ToRadians(Rotation));

            // set flip properties
            _quad.FlipX = FlipX;
            _quad.FlipY = FlipY;

            // refill vertex buffer with Material's region coordinates
            // (note: this must happen after flip is assigned for flip properties to apply properly)
            if (_quad.RefillVB && Material != null)
                _quad.TextureCoordinates = Material.GetRegionCoords(MaterialRegionIndex);

            objToWorld = ScaleMatrix * RotationMatrix * TranslationMatrix;

            // forward to quad
            _quad.Render(objToWorld, VisibilityLevel, srs);

            base.Render(srs);

            Profiler.Instance.EndBlock("T2DStaticSprite.Render");
        }

        public override void CopyTo(TorqueObject obj)
        {
            base.CopyTo(obj);

            T2DStaticSprite obj2 = (T2DStaticSprite)obj;
            obj2.Material = Material;
        }

        public void Dispose()
        {
            _quad.Dispose();
        }

        public void UpdateVertexBuffer()
        {
            if (Material == null)
                return;

            _quad.TextureCoordinates = Material.GetRegionCoords(MaterialRegionIndex);

            _quad.RefillVB = true;

        }

        #endregion

        //======================================================
        #region Private, protected, internal fields
        RenderQuad _quad = new RenderQuad();
        protected int _materialRegionIndex;
        #endregion
    }
}
