//-----------------------------------------------------------------------------
// Torque X Game Engine
// Copyright  GarageGames.com, Inc.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using GarageGames.Torque.Core;
using GarageGames.Torque.GFX;
using GarageGames.Torque.Materials;
using GarageGames.Torque.MathUtil;
using GarageGames.Torque.SceneGraph;
using GarageGames.Torque.XNA;
using GarageGames.Torque.Util;
using GarageGames.Torque.RenderManager;

namespace GarageGames.Torque.GUI
{
    /// <summary>
    /// DrawUtil provides convenient 2D drawing capabilities used by the GUI.
    /// </summary>
    public class DrawUtil
    {
        //======================================================
        #region Static methods, fields, constructors

        /// <summary>
        /// Sets up the draw util.
        /// </summary>
        public static void Setup()
        {
            // set up the default gui material here
            _defaultMaterial = new Materials.SimpleMaterial();
            _defaultMaterial.TextureFilename = null;
            _defaultMaterial.IsTranslucent = true;

            // set up the default render instance to be used when rendering rects
            _rectRenderInstance = new RenderInstance();
            _rectRenderInstance.ObjectTransform = Matrix.Identity;

            // set up the vertex buffer
            _CreateVertexBuffer();
        }

        /// <summary>
        /// Sets up an orthographical viewport and clipping region.
        /// </summary>
        public static RectangleF ClipRect
        {
            get { return _clipRect; }
            set
            {
                _clipRect = value;

                float left = _clipRect.X;
                float right = _clipRect.X + _clipRect.Width;
                float bottom = _clipRect.Y + _clipRect.Height;
                float top = _clipRect.Y;

                // setup projection matrix
                _clipMatrix = GFXDevice.Instance.SetOrtho(false, left, right, bottom, top, 0.0f, 1.0f);

                Viewport viewport = new Viewport();
                viewport.X = (int)_clipRect.X;
                viewport.Y = (int)_clipRect.Y;
                viewport.Width = (int)_clipRect.Width;
                viewport.Height = (int)_clipRect.Height;
                viewport.MinDepth = 0.0f;
                viewport.MaxDepth = 1.0f;

                GFXDevice.Instance.Device.Viewport = viewport;
            }
        }

        /// <summary>
        /// Modulates a texture's color value based on the given modulation. This can
        /// be used to 'tint' a texture to a specific color.
        /// </summary>
        public static Color BitmapModulation
        {
            get { return _bitmapModulation; }
            set { _bitmapModulation = value; }
        }

        /// <summary>
        /// Resets the modulated texture color value to normal.
        /// </summary>
        public static void ClearBitmapModulation()
        {
            _bitmapModulation = new Color(255, 255, 255, 255);
        }

        /// <summary>
        /// Draws an untextured filled rectangle from the upper left, <paramref name="a"/>,
        /// of the rectangle to the lower right, <paramref name="b"/> in the specified
        /// color, <paramref name="color"/>.
        /// </summary>
        public static void RectFill(Vector2 a, Vector2 b, Color color)
        {
            //
            // Convert Box   a----------x
            //               |          |
            //               x----------b
            // Into Quad
            //               v0---------v1
            //               | a       x |
            //			     |           |
            //               | x       b |
            //               v2---------v3
            //

            SceneRenderState srs = GUICanvas.Instance.RenderState;
            GFXDevice gfx = srs.Gfx;

            // setup render state
            srs.World.LoadIdentity();
            srs.View = Matrix.Identity;
            srs.Projection = _clipMatrix;

            // setup the material
            _defaultMaterial.SetupEffect(srs, null);


            Vector2 nw = new Vector2(-0.5f, -0.5f);
            Vector2 ne = new Vector2(+0.5f, -0.5f);

            GFXVertexFormat.PCTTBN[] verts = TorqueUtil.GetScratchArray<GFXVertexFormat.PCTTBN>(4);

            verts[0] = new GFXVertexFormat.PCTTBN();
            verts[0].Position = new Vector3(a.X + nw.X, a.Y + nw.Y, 0.0f);
            verts[0].Color = color;

            verts[1] = new GFXVertexFormat.PCTTBN();
            verts[1].Position = new Vector3(b.X + ne.X, a.Y + ne.Y, 0.0f);
            verts[1].Color = color;

            verts[2] = new GFXVertexFormat.PCTTBN();
            verts[2].Position = new Vector3(a.X - ne.X, b.Y - ne.Y, 0.0f);
            verts[2].Color = color;

            verts[3] = new GFXVertexFormat.PCTTBN();
            verts[3].Position = new Vector3(b.X - nw.X, b.Y - nw.Y, 0.0f);
            verts[3].Color = color;

            // place the verts in the vertex buffer
            if (_vertexBuffer.IsNull)
                _CreateVertexBuffer();

            // set up the material
            _defaultMaterial.SetupObject(_rectRenderInstance, srs);

            // draw the vertices
            while (_defaultMaterial.SetupPass())
                gfx.Device.DrawUserPrimitives<GFXVertexFormat.PCTTBN>(PrimitiveType.TriangleStrip, verts, 0, 2);

            // clean up the material
            _defaultMaterial.CleanupEffect();
        }

        /// <summary>
        /// Draws an untextured filled rectangle from <paramref name="rect"/> in the specified
        /// color, <paramref name="color"/>.
        /// </summary>
        public static void RectFill(RectangleF rect, Color color)
        {
            Vector2 lowerRight = new Vector2(rect.Width + rect.X - 1, rect.Height + rect.Y - 1);
            DrawUtil.RectFill(rect.Point, lowerRight, color);
        }

        /// <summary>
        /// Draws an untextured wireframe rectangle from the upper left, <paramref name="a"/>,
        /// of the rectangle to the lower right, <paramref name="b"/> in the specified
        /// color, <paramref name="color"/>.
        /// </summary>
        public static void Rect(Vector2 a, Vector2 b, Color color)
        {
            //
            // Convert Box   a----------x
            //               |          |
            //               x----------b
            //
            // Into tri-Strip Outline
            //               v2-----------v0
            //               | a         x |
            //               |  v1-----v7  |
            //               |   |     |   |
            //               |  v3-----v5  |
            //               | x         b |
            //               v4-----------v6
            //


            SceneRenderState srs = GUICanvas.Instance.RenderState;
            GFXDevice gfx = srs.Gfx;

            // setup render state
            srs.World.LoadIdentity();
            srs.View = Matrix.Identity;
            srs.Projection = _clipMatrix;

            // setup the material
            _defaultMaterial.SetupEffect(srs, null);


            Vector2 nw = new Vector2(-0.5f, -0.5f);
            Vector2 ne = new Vector2(+0.5f, -0.5f);

            GFXVertexFormat.PCTTBN[] verts = TorqueUtil.GetScratchArray<GFXVertexFormat.PCTTBN>(10);

            verts[0] = new GFXVertexFormat.PCTTBN();
            verts[0].Position = new Vector3(b.X + ne.X, a.Y + ne.Y, 0.0f);
            verts[0].Color = color;

            verts[1] = new GFXVertexFormat.PCTTBN();
            verts[1].Position = new Vector3(a.X - nw.X, a.Y - nw.Y, 0.0f);
            verts[1].Color = color;

            verts[2] = new GFXVertexFormat.PCTTBN();
            verts[2].Position = new Vector3(a.X + nw.X, a.Y + nw.Y, 0.0f);
            verts[2].Color = color;

            verts[3] = new GFXVertexFormat.PCTTBN();
            verts[3].Position = new Vector3(a.X + ne.X, b.Y + ne.Y, 0.0f);
            verts[3].Color = color;

            verts[4] = new GFXVertexFormat.PCTTBN();
            verts[4].Position = new Vector3(a.X - ne.X, b.Y - ne.Y, 0.0f);
            verts[4].Color = color;

            verts[5] = new GFXVertexFormat.PCTTBN();
            verts[5].Position = new Vector3(b.X + nw.X, b.Y + nw.Y, 0.0f);
            verts[5].Color = color;

            verts[6] = new GFXVertexFormat.PCTTBN();
            verts[6].Position = new Vector3(b.X - nw.X, b.Y - nw.Y, 0.0f);
            verts[6].Color = color;

            verts[7] = new GFXVertexFormat.PCTTBN();
            verts[7].Position = new Vector3(b.X - ne.X, a.Y - ne.Y, 0.0f);
            verts[7].Color = color;

            verts[8] = new GFXVertexFormat.PCTTBN();
            verts[8].Position = new Vector3(b.X + ne.X, a.Y + ne.Y, 0.0f); // same as v0
            verts[8].Color = color;

            verts[9] = new GFXVertexFormat.PCTTBN();
            verts[9].Position = new Vector3(a.X - nw.X, a.Y - nw.Y, 0.0f); // same as v1
            verts[9].Color = color;

            // place the verts in the vertex buffer
            if (_vertexBuffer.IsNull)
                _CreateVertexBuffer();

            gfx.Device.VertexDeclaration = GFXVertexFormat.GetVertexDeclaration(gfx.Device);
            gfx.Device.DrawUserPrimitives<GFXVertexFormat.PCTTBN>(PrimitiveType.TriangleStrip, verts, 0, 2);

            // set up the material
            _defaultMaterial.SetupObject(_rectRenderInstance, srs);

            // draw the vertices
            while (_defaultMaterial.SetupPass())
                gfx.Device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 8);
            //                GFXDevice.Instance.Device.DrawUserPrimitives<GFXVertexFormat.PCTTBN>(PrimitiveType.TriangleStrip, verts, 0, 9);

            // cleanup the material
            _defaultMaterial.CleanupEffect();
        }

        /// <summary>
        /// Draws an untextured wireframe rectangle from <paramref name="rect"/> in the specified
        /// color, <paramref name="color"/>.
        /// </summary>
        public static void Rect(RectangleF rect, Color color)
        {
            Vector2 lowerRight = new Vector2(rect.Width + rect.X - 1, rect.Height + rect.Y - 1);
            DrawUtil.Rect(rect.Point, lowerRight, color);
        }

        /// <summary>
        /// Draws a stretched sub-region of a texture.
        /// </summary>
        /// <param name="material">The material used when rendering. The material must implement ITextureMaterial.</param>
        /// <param name="dstRect">Rectangle where the texture object will be drawn.</param>
        /// <param name="srcRect">Sub-region of the texture that will be applied over the <paramref name="dstRect"/>.</param>
        /// <param name="flipMode">Any flipping to be done of the source texture.</param>
        public static void BitmapStretchSR(RenderMaterial material, RectangleF dstRect, RectangleF srcRect, BitmapFlip flipMode)
        {
            SceneRenderState srs = GUICanvas.Instance.RenderState;
            GFXDevice gfx = srs.Gfx;

            // setup render state
            srs.World.LoadIdentity();
            srs.View = Matrix.Identity;
            srs.Projection = _clipMatrix;

            // setup the material
            material.SetupEffect(srs, null);

            GFXVertexFormat.PCTTBN[] verts = TorqueUtil.GetScratchArray<GFXVertexFormat.PCTTBN>(4);

            // from the source rectangle, determine the texture coordinates
            Texture2D tex = (Texture2D)(material as ITextureMaterial).Texture.Instance;
            float texLeft = srcRect.X / tex.Width;
            float texRight = (srcRect.X + srcRect.Width) / tex.Width;
            float texTop = srcRect.Y / tex.Height;
            float texBottom = (srcRect.Y + srcRect.Height) / tex.Height;

            float screenLeft = dstRect.X;
            float screenRight = dstRect.X + dstRect.Width;
            float screenTop = dstRect.Y;
            float screenBottom = dstRect.Y + dstRect.Height;

            // flip x
            if ((flipMode & BitmapFlip.FlipX) != 0)
            {
                float temp = texLeft;
                texLeft = texRight;
                texRight = temp;
            }

            // flip y
            if ((flipMode & BitmapFlip.FlipY) != 0)
            {
                float temp = texTop;
                texTop = texBottom;
                texBottom = temp;
            }

            Color color = _bitmapModulation;

            verts[0] = new GFXVertexFormat.PCTTBN();
            verts[0].Position = new Vector3(screenLeft - 0.5f, screenTop - 0.5f, 0.0f);
            verts[0].TextureCoordinate = new Vector2(texLeft, texTop);
            verts[0].Color = color;

            verts[1] = new GFXVertexFormat.PCTTBN();
            verts[1].Position = new Vector3(screenRight - 0.5f, screenTop - 0.5f, 0.0f);
            verts[1].TextureCoordinate = new Vector2(texRight, texTop);
            verts[1].Color = color;

            verts[2] = new GFXVertexFormat.PCTTBN();
            verts[2].Position = new Vector3(screenLeft - 0.5f, screenBottom - 0.5f, 0.0f);
            verts[2].TextureCoordinate = new Vector2(texLeft, texBottom);
            verts[2].Color = color;

            verts[3] = new GFXVertexFormat.PCTTBN();
            verts[3].Position = new Vector3(screenRight - 0.5f, screenBottom - 0.5f, 0.0f);
            verts[3].TextureCoordinate = new Vector2(texRight, texBottom);
            verts[3].Color = color;

            // place the verts in the vertex buffer
            if (_vertexBuffer.IsNull)
                _CreateVertexBuffer();

            // adltodo: hacks
            RenderInstance ri = SceneRenderer.RenderManager.AllocateInstance();
            ri.ObjectTransform = Matrix.Identity;

            // draw the vertices
            while (material.SetupPass())
            {
                material.SetupObject(ri, srs);

                gfx.Device.VertexDeclaration = GFXVertexFormat.GetVertexDeclaration(gfx.Device);
                gfx.Device.DrawUserPrimitives<GFXVertexFormat.PCTTBN>(PrimitiveType.TriangleStrip, verts, 0, 2);
            }

            // cleanup the material
            material.CleanupEffect();
        }

        /// <summary>
        /// Draws an unstretched bitmap.
        /// </summary>
        /// <param name="material">The material used when rendering. The material must implement ITextureMaterial.</param>
        /// <param name="position">Where to draw the texture in 2d coordinates.</param>
        /// <param name="flipMode">Any flipping to be done of the source texture.</param>
        public static void Bitmap(RenderMaterial material, Vector2 position, BitmapFlip flipMode)
        {
            Assert.Fatal(material != null, "No material specified for DrawUtil::Bitmap");

            ///////////////////////////////////////////////////////////////////
            // rdbnote: the render material should really handle this for me
            Resource<Texture> resource;
            ITextureMaterial iMat = material as ITextureMaterial;
            bool isInvalid = iMat.Texture.IsNull;

            if (!isInvalid)
                resource = iMat.Texture;
            else
                resource = ResourceManager.Instance.LoadTexture((material as ITextureMaterial).TextureFilename);

            Texture2D texture = (Texture2D)resource.Instance;
            ///////////////////////////////////////////////////////////////////

            RectangleF subRegion = new RectangleF(0.0f, 0.0f, texture.Width, texture.Height);
            RectangleF stretch = new RectangleF(position.X, position.Y, texture.Width, texture.Height);
            DrawUtil.BitmapStretchSR(material, stretch, subRegion, flipMode);

            // invalidate the temp. resource
            if (isInvalid)
                resource.Invalidate();
        }

        /// <summary>
        /// Draws a stretched bitmap.
        /// </summary>
        /// <param name="material">The material used when rendering. The material must implement ITextureMaterial.</param>
        /// <param name="dstRect">Rectangle where the texture object will be drawn.</param>
        /// <param name="flipMode">Any flipping to be done of the source texture.</param>
        public static void BitmapStretch(RenderMaterial material, RectangleF dstRect, BitmapFlip flipMode)
        {
            Assert.Fatal(material != null, "No material specified for DrawUtil::BitmapStretch");

            ///////////////////////////////////////////////////////////////////
            // rdbnote: the render material should really handle this for me
            Resource<Texture> resource;
            ITextureMaterial iMat = material as ITextureMaterial;

            Assert.Fatal(iMat != null, "BitmapStretch only works on ITextureMaterial!");

            bool isInvalid = iMat.Texture.IsNull;

            if (!isInvalid)
                resource = iMat.Texture;
            else
                resource = ResourceManager.Instance.LoadTexture((material as ITextureMaterial).TextureFilename);

            Texture2D texture = (Texture2D)resource.Instance;
            ///////////////////////////////////////////////////////////////////

            RectangleF subRegion = new RectangleF(0.0f, 0.0f, texture.Width, texture.Height);
            DrawUtil.BitmapStretchSR(material, dstRect, subRegion, flipMode);

            // invalidate the temp. resource
            if (isInvalid)
                resource.Invalidate();
        }

        /// <summary>
        /// Draws an unstretched sub-region of a texture.
        /// </summary>
        /// <param name="material">The material used when rendering. The material must implement ITextureMaterial.</param>
        /// <param name="position">Where to draw the texture in 2d coordinates.</param>
        /// <param name="srcRect">Sub-region of the texture to be drawn.</param>
        /// <param name="flipMode">Any flipping to be done of the source texture.</param>
        public static void BitmapSR(RenderMaterial material, Vector2 position, RectangleF srcRect, BitmapFlip flipMode)
        {
            Assert.Fatal(material != null, "No texture specified for DrawUtil::BitmapSR");

            RectangleF stretch = new RectangleF(position.X, position.Y, srcRect.Width, srcRect.Height);
            DrawUtil.BitmapStretchSR(material, stretch, srcRect, flipMode);
        }

        /// <summary>
        /// Deaws text at a location in the 2d gui coordinates.
        /// </summary>
        /// <param name="font">The font to draw with, usually specified by a GUIStyle.</param>
        /// <param name="offset">Point where to start drawing the text.</param>
        /// <param name="size">Margins used to calculate the text alignment.</param>
        /// <param name="alignment">Justification of the text relative to the offset.</param>
        /// <param name="color">The color to draw the text as.</param>
        /// <param name="text">The string to be drawn.</param>
        public static void JustifiedText(Resource<SpriteFont> font, Vector2 offset, Vector2 size, TextAlignment alignment, Color color, string text)
        {
            if (font.IsNull)
                return;

            float textWidth = font.Instance.MeasureString(text).X;
            Vector2 start;

            switch (alignment)
            {
                case TextAlignment.JustifyRight:
                    start = new Vector2(size.X - textWidth, 0.0f);
                    break;
                case TextAlignment.JustifyCenter:
                    start = new Vector2((size.X - textWidth) / 2.0f, 0.0f);
                    break;
                default:
                    start = new Vector2(0.0f, 0.0f);
                    break;
            }

            // if the text is longer then the box size it will get clipped,
            // so force left justify
            if (textWidth > size.X)
                start = new Vector2(0.0f, 0.0f);

            // center vertical
            if (font.Instance.LineSpacing > size.Y)
                start.Y = 0.0f - ((font.Instance.LineSpacing - size.Y) / 2.0f);
            else
                start.Y = (size.Y - font.Instance.LineSpacing) / 2.0f;

            FontRenderer.Instance.DrawString(font, start + offset, color, text);
        }

        private static void _CreateVertexBuffer()
        {
            int sizeInBytes = 1024 * GFXVertexFormat.VertexSize;
            _vertexBuffer = ResourceManager.Instance.CreateVertexBuffer(ResourceProfiles.ManualStaticVBProfile, sizeInBytes);

            Assert.Fatal(!_vertexBuffer.IsNull, "DrawUtil::_CreateVertexBuffer: failed to create a vertex buffer!");
        }

        static RenderInstance _rectRenderInstance;
        static RectangleF _clipRect;
        static Matrix _clipMatrix;
        static Color _bitmapModulation = new Color(255, 255, 255, 255);
        static Materials.SimpleMaterial _defaultMaterial;
        static Resource<DynamicVertexBuffer> _vertexBuffer;

        #endregion
    }
}
