//-----------------------------------------------------------------------------
// Torque X Game Engine
// Copyright  GarageGames.com, Inc.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using GarageGames.Torque.GFX;
using GarageGames.Torque.Materials;
using GarageGames.Torque.SceneGraph;
using GarageGames.Torque.Util;
using GarageGames.Torque.XNA;
using GarageGames.Torque.Core;

namespace GarageGames.Torque.RenderManager
{
    /// <summary>
    /// Manages rendering materials that refract the scene beneath them. In 3D this works pretty much as expected, but
    /// in 2D there are definitely some caveats. By default a 2D scene is not rendered with a depth buffer, so, because
    /// the refraction manager is rendered last, everything in the scene at the same location as the refraction object
    /// will be refracted. You can turn on depth buffering for a 2D scene (T2DSceneGraph.UseDepthBuffer), which generally
    /// works, but requires you to be smart about which objects are rendered as type Mesh2D and which are rendered as type
    /// Translucent (in other words, the IsTranslucent flag on RenderMaterial must be set properly). Still, everything in
    /// the translucent render manager will be refracted if it is in the same location as the refracting object because
    /// translucent objects don't render to the depth buffer.
    /// 
    /// Refraction is handled by resolving the back buffer and copying it to a texture. That texture is set on a material
    /// (RefractionMaterial, unless a custom material is created that refracts) before objects are rendered.
    /// </summary>
    public class RefractionManager : BaseRenderManager
    {
        //======================================================
        #region Public methods

        protected override void _RenderGroup(List<RenderInstance> renderInstances, RenderMaterial material, MaterialInstanceData materialData, SceneRenderState srs, GraphicsDevice d3d)
        {
            (material as IRefractionMaterial).SetTexture(_textureRenderEffect);
            base._RenderGroup(renderInstances, material, materialData, srs, d3d);
        }

        public override void RenderZPass(SceneRenderState srs)
        {
        }

        public override void RenderOpaquePass(SceneRenderState srs)
        {
        }

        public override void RenderTranslucentPass(SceneRenderState srs)
        {
            if (_elementList.Count == 0 || GFXDevice.Instance.ShaderProfile < ShaderProfile.PS_1_4 ||
                !TorqueEngineComponent.Instance.EnableBackBufferEffects || srs.IsReflectPass ||
                TorqueEngineComponent.Instance.CurrentRenderTarget == null)
            {
                return;
            }

            Profiler.Instance.StartBlock(_renderProfileBlock);

            RenderTarget2D rendertargetScene = GFXDevice.Instance.Device.GetRenderTarget(0) as RenderTarget2D;

            Vector2 size = new Vector2(
                TorqueEngineComponent.Instance.CurrentRenderTarget.Width,
                TorqueEngineComponent.Instance.CurrentRenderTarget.Height);

            if (_rendertargetEffect.IsNull || _rendertargetEffect.IsInvalid)
                _rendertargetEffect = ResourceManager.Instance.CreateRenderTarget2D((int)size.X, (int)size.Y, 1, SurfaceFormat.Color);

            GFXDevice.Instance.Device.SetRenderTarget(0, _rendertargetEffect.Instance);

            if (rendertargetScene != null && rendertargetScene is RenderTarget2D)
                _textureRenderScene = rendertargetScene.GetTexture();

            if (_copyMaterial == null)
            {
                _copyMaterial = new SimpleMaterial();
                _copyMaterial.IsCopyPass = true;
            }

            _copyMaterial.SetTexture(_textureRenderScene);

            SceneRenderer.RenderManager.RenderQuad(_copyMaterial, Vector2.Zero, size);

            GFXDevice.Instance.Device.SetRenderTarget(0, rendertargetScene);

            _textureRenderEffect = _rendertargetEffect.Instance.GetTexture();

            TorqueEngineComponent.Instance.ReapplyMainRenderTarget();
            base.RenderOpaquePass(srs);

            Profiler.Instance.EndBlock(_renderProfileBlock);
        }

        #endregion

        //======================================================
        #region Private, protected, internal fields

        Texture2D _textureRenderScene;
        Texture2D _textureRenderEffect;
        Resource<RenderTarget2D> _rendertargetEffect;
        SimpleMaterial _copyMaterial;

        ProfilerCodeBlock _renderProfileBlock = new ProfilerCodeBlock("RefractionManager.RenderTranslucentPass");

        #endregion
    }
}
