//-----------------------------------------------------------------------------
// Torque X Game Engine
// Copyright  GarageGames.com, Inc.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using GarageGames.Torque.Core;
using GarageGames.Torque.GFX;
using GarageGames.Torque.Materials;
using GarageGames.Torque.MathUtil;
using GarageGames.Torque.SceneGraph;
using GarageGames.Torque.Sim;
using GarageGames.Torque.RenderManager;
using GarageGames.Torque.Util;

namespace GarageGames.Torque.RenderManager
{
    public class RenderQuad : IDisposable
    {
        //======================================================
        #region Constructors

        public RenderQuad()
        {
            _numVerts = 4;
        }

        #endregion

        //======================================================
        #region Public properties, operators, constants, and enums

        public RenderMaterial Material
        {
            set { _material = value; }
            get { return _material; }
        }

        public Vector2 Size
        {
            set { _size = value; Points = null; _refillVB = true; }
        }

        // used to set a 4 element vector of points for the quad.
        // if you set this, the specified points will be used to draw the quad rather.  Otherwise
        // the quad points are computed based on the Size parameter (see MakePoints) below
        // quad winding is as follows:
        //  3--2
        //  |  |
        //  0--1
        public Vector2[] Points
        {
            set { _points = value; _refillVB = true; Assert.Fatal(_points == null || _points.Length == _numVerts, "doh"); }
            get { return _points; }
        }

        public TextureAddressMode TexAddressU
        {
            set { _texU = value; }
            get { return _texU; }
        }

        public TextureAddressMode TexAddressV
        {
            set { _texV = value; }
            get { return _texV; }
        }

        public Vector2[] TextureCoords
        {
            set { _textureCoords = value; _refillVB = true; }
            get { return _textureCoords; }
        }

        public float TextureOffsetX
        {
            set { _textureOffset.X = value; _refillVB = true; }
            get { return _textureOffset.X; }
        }

        public float TextureOffsetY
        {
            set { _textureOffset.Y = value; _refillVB = true; }
            get { return _textureOffset.Y; }
        }

        public float TextureRepeatX
        {
            set { _textureRepeat.X = value; _refillVB = true; }
            get { return _textureRepeat.X; }
        }

        public float TextureRepeatY
        {
            set { _textureRepeat.Y = value; _refillVB = true; }
            get { return _textureRepeat.Y; }
        }

        /// <summary>
        /// Flips the quad's texture coordinates horizontally
        /// </summary>
        public bool FlipX
        {
            set { if (_flipX == value) return; _flipX = value; _refillVB = true; }
            get { return _flipX; }
        }

        /// <summary>
        /// Flips the quad's texture coordinates vertically
        /// </summary>
        public bool FlipY
        {
            set { if (_flipY == value) return; _flipY = value; _refillVB = true; }
            get { return _flipY; }
        }

        /// <summary>
        /// Specify this as TRUE to not automatically compute new texture coords.
        /// </summary>
        public bool IgnoreTextureCoordsUpdate
        {
            get { return _ignoreTextureCoordsUpdate; }
            set { _ignoreTextureCoordsUpdate = value; _refillVB = true; }
        }

        /// <summary>
        /// Get/Set the current texture area that is being displayed on the quad/sprite.
        /// <para>The texture area's physical bounds are 0-1 for both X and Y.</para>
        /// <para>Values outside of 0-1 lead to texture 'wrapping'.</para>
        /// </summary>
        public RectangleF TextureCoordinates
        {
            get
            {
                float x = _textureCoords[0].X;
                float y = _textureCoords[0].Y;
                float width = _textureCoords[2].X - x;
                float height = _textureCoords[2].Y - y;
                return new RectangleF(x, y, width, height);
            }
            set
            {
                float minX = value.X;
                float minY = value.Y;
                float maxX = minX + value.Width;
                float maxY = minY + value.Height;
                _SetTextureCoords(minX, minY, maxX, maxY);
                _refillVB = true;
            }
        }

        public bool RefillVB
        {
            get { return _refillVB; }
            set { _refillVB = value; }
        }

        #endregion

        //======================================================
        #region Public methods

        // populate the object.
        public void CreateAndFillVB()
        {
            if (_vb == null)
                _CreateVB();

            Assert.Fatal((_vb != null), "Invalid buffer.");

            if (_refillVB || !_vb.AreBufferContentsValid())
                _FillVB();
        }

        public void FillRenderInstance(RenderInstance ri, float opacity, SceneRenderState srs)
        {
            GraphicsDevice d3d = srs.Gfx.Device;
            Assert.Fatal(d3d != null, "doh");

            ri.Type = RenderInstance.RenderInstanceType.Mesh2D;

            ri.ObjectTransform = srs.World.Top;

            ri.PrimitiveType = PrimitiveType.TriangleFan;
            ri.VertexDeclaration = GFXVertexFormat.GetVertexDeclaration(srs.Gfx.Device);

            ri.VertexBuffer = _vb.Buffer as VertexBuffer;
            ri.BaseVertex = _vb.StartIndex;
            ri.VertexCount = _vb.Count;
            ri.VertexSize = _vb.ElementSize;

            ri.PrimitiveCount = 2;


            ri.Opacity = opacity;
            ri.UTextureAddressMode = _texU;
            ri.VTextureAddressMode = _texV;

            ri.Material = _material;
        }

        // render the object.
        public void Render(Matrix objToWorld, float opacity, SceneRenderState srs)
        {
            CreateAndFillVB();

            Assert.Fatal(_vb != null, "doh");
            Assert.Fatal(!_refillVB, "doh, vb needs to be refilled");

            srs.World.Push();
            srs.World.MultiplyMatrixLocal(objToWorld);

            RenderInstance ri = SceneRenderer.RenderManager.AllocateInstance();

            FillRenderInstance(ri, opacity, srs);

            SceneRenderer.RenderManager.AddInstance(ri);
            srs.World.Pop();
        }

        public void Dispose()
        {
            if (_vb != null)
            {
                _vb.Dispose();
                _vb = null;
            }
        }

        #endregion

        //======================================================
        #region Private, protected, internal methods

        private void _CreatePoints()
        {
            _points = new Vector2[4];

            float minX = -1.0f;
            float minY = -1.0f;
            float maxX = 1.0f;
            float maxY = 1.0f;

            _points[0] = new Vector2(minX, minY);
            _points[1] = new Vector2(maxX, minY);
            _points[2] = new Vector2(maxX, maxY);
            _points[3] = new Vector2(minX, maxY);
        }

        private void _CreateTangentSpace(out Vector4 normal, out Vector4 tangent, out Vector4 binormal)
        {
            Vector3 p0 = new Vector3(_points[0].X, _points[0].Y, 0.0f);
            Vector3 p1 = new Vector3(_points[1].X, _points[1].Y, 0.0f);
            Vector3 p2 = new Vector3(_points[2].X, _points[2].Y, 0.0f);
            Vector3 p3 = new Vector3(_points[3].X, _points[3].Y, 0.0f);

            Vector3 v3n = Vector3.Cross(p1 - p0, p2 - p0);
            normal = new Vector4(v3n.X, v3n.Y, v3n.Z, 0.0f);
            if (normal.LengthSquared() > 0.0001f)
                normal.Normalize();

            Vector3 v3tangent = p1 - p0;
            tangent = new Vector4(v3tangent.X, v3tangent.Y, v3tangent.Z, 0.0f);
            if (tangent.LengthSquared() > 0.0001f)
                tangent.Normalize();

            Vector3 v3binormal = p3 - p0;
            binormal = new Vector4(v3binormal.X, v3binormal.Y, v3binormal.Z, 0.0f);
            if (binormal.LengthSquared() > 0.0001f)
                binormal.Normalize();
        }

        private void _CreateTextureCoords()
        {
            float minX = 0.00f;
            float minY = 0.00f;
            float maxX = 1.0f;
            float maxY = 1.0f;

            minX = (minX + _textureOffset.X);
            minY = (minY + _textureOffset.Y);
            maxX = (maxX * _textureRepeat.X + _textureOffset.X);
            maxY = (maxY * _textureRepeat.Y + _textureOffset.Y);
            _SetTextureCoords(minX, minY, maxX, maxY);
        }


        private void _SetTextureCoords(float minX, float minY, float maxX, float maxY)
        {
            _textureCoords[0] = new Vector2(minX, minY);
            _textureCoords[1] = new Vector2(maxX, minY);
            _textureCoords[2] = new Vector2(maxX, maxY);
            _textureCoords[3] = new Vector2(minX, maxY);

            if (_flipX)
            {
                TorqueUtil.Swap<Vector2>(ref _textureCoords[0], ref _textureCoords[1]);
                TorqueUtil.Swap<Vector2>(ref _textureCoords[2], ref _textureCoords[3]);
            }
            if (_flipY)
            {
                TorqueUtil.Swap<Vector2>(ref _textureCoords[0], ref _textureCoords[3]);
                TorqueUtil.Swap<Vector2>(ref _textureCoords[1], ref _textureCoords[2]);
            }
        }

        private void _CreateVB()
        {
            if (_vb != null)
                return;

            Assert.Fatal(_vb == null, "doh");

            _vb = new GFXDynamicVertexBufferPCTTBN(_numVerts);
            //_vb = new GFXVolatileSharedVertexBufferPCTTBN(_numVerts);

            _refillVB = true;
        }

        private void _FillVB()
        {
            Assert.Fatal(_vb != null, "doh");
            Assert.Fatal(_numVerts > 0, "doh");

            if (_points == null)
                _CreatePoints();

            if (!_ignoreTextureCoordsUpdate)
            {
                _CreateTextureCoords();
            }

            Vector4 normal, tangent, binormal;
            _CreateTangentSpace(out normal, out tangent, out binormal);

            Color color = Color.White;

            // fill in vertex array 
            int sizeInBytes = _numVerts * GFXVertexFormat.VertexSize;
            GFXVertexFormat.PCTTBN[] vertices = _vb.GetScratchArray(_numVerts);
            for (int i = 0; i < _numVerts; ++i)
            {
                vertices[i] = new GFXVertexFormat.PCTTBN(
                    new Vector3(_points[i].X, Points[i].Y, 0.0f),
                    color,
                    _textureCoords[i],
                    _textureCoords[i],
                    tangent,
                    normal);
            }

            _vb.SetData(vertices, 0, _numVerts);

            _refillVB = false;
        }

        #endregion

        //======================================================
        #region Private, protected, internal fields
        private IGFXBuffer<GFXVertexFormat.PCTTBN> _vb;
        private RenderMaterial _material;
        private Vector2 _size;
        private Vector2[] _points;
        private Vector2[] _textureCoords = new Vector2[4];
        private Vector2 _textureOffset = new Vector2(0.0f, 0.0f);
        private Vector2 _textureRepeat = new Vector2(1.0f, 1.0f);
        private TextureAddressMode _texU = TextureAddressMode.Clamp;
        private TextureAddressMode _texV = TextureAddressMode.Clamp;
        private bool _flipX;
        private bool _flipY;
        private bool _refillVB;
        private int _numVerts;
        private bool _ignoreTextureCoordsUpdate;
        #endregion
    }
}
