//-----------------------------------------------------------------------------
// Torque X Game Engine
// Copyright  GarageGames.com, Inc.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using GarageGames.Torque.Materials;
using GarageGames.Torque.RenderManager;
using GarageGames.Torque.SceneGraph;
using GarageGames.Torque.Sim;
using GarageGames.Torque.Core;
using GarageGames.Torque.Util;

namespace GarageGames.Torque.T2D
{
    public class T2DScroller : T2DSceneObject, IAnimatedObject, IDisposable
    {
        //======================================================
        #region Constructors

        public T2DScroller()
        {
            // disable collisions on scrollers by default
            CollisionsEnabled = false;
        }

        #endregion

        //======================================================
        #region Public properties, operators, constants, and enums

        /// <summary>
        /// Material of the scroller
        /// </summary>
        public RenderMaterial Material
        {
            set { _q.Material = value; }
            get { return _q.Material; }
        }
        /// <summary>
        /// Changes the rate the scroller moves horizontally.
        /// </summary>
        public float ScrollRateX
        {
            set { _scrollX = value; }
            get { return _scrollX; }
        }
        /// <summary>
        /// Changes the rate the scroller moves vertically.
        /// </summary>
        public float ScrollRateY
        {
            set { _scrollY = value; }
            get { return _scrollY; }
        }
        /// <summary>
        /// Specifies the X offset of the scroller texture 
        /// </summary>
        public float TextureOffsetX
        {
            set { _q.TextureOffsetX = value; }
            get { return _q.TextureOffsetX; }
        }
        /// <summary>
        /// Specifies the Y offset of the scroller texture 
        /// </summary>
        public float TextureOffsetY
        {
            set { _q.TextureOffsetY = value; }
            get { return _q.TextureOffsetY; }
        }
        /// <summary>
        /// Specifies how many times the texture repeats horizontally accross the scroll area.
        /// </summary>
        public float TextureRepeatX
        {
            set { _q.TextureRepeatX = value; }
            get { return _q.TextureRepeatX; }
        }
        /// <summary>
        /// Specifies how many times the texture repeats vertically accross the scroll area.
        /// </summary>
        public float TextureRepeatY
        {
            set { _q.TextureRepeatY = value; }
            get { return _q.TextureRepeatY; }
        }

        public override string AutoName
        {
            get
            {
                string autoname = base.AutoName;
                // add material name
                if (Material != null)
                    autoname += "_" + Material.AutoName;

                return autoname;
            }
        }

        #endregion

        //======================================================
        #region Public methods

        public override bool OnRegister()
        {
            if (!base.OnRegister())
                return false;

            ProcessList.Instance.AddAnimationCallback(this);
            return true;
        }

        public override void CopyTo(TorqueObject obj)
        {
            base.CopyTo(obj);
            T2DScroller obj2 = (T2DScroller)obj;
            obj2.Material = Material;
            obj2.ScrollRateX = ScrollRateX;
            obj2.ScrollRateY = ScrollRateY;
            obj2.TextureOffsetX = TextureOffsetX;
            obj2.TextureOffsetY = TextureOffsetY;
            obj2.TextureRepeatX = TextureRepeatX;
            obj2.TextureRepeatY = TextureRepeatY;
        }

        public override void Render(SceneRenderState srs)
        {
            Profiler.Instance.StartBlock("T2DScroller.Render");

            Matrix ScaleMatrix = Matrix.CreateScale(0.5f * Size.X, 0.5f * Size.Y, 1.0f);
            Matrix TranslationMatrix = Matrix.CreateTranslation(Position.X, Position.Y, LayerDepth);
            Matrix RotationMatrix = Matrix.CreateRotationZ(MathHelper.ToRadians(Rotation));
            Matrix ObjToWorld = ScaleMatrix * RotationMatrix * TranslationMatrix;

            _q.FlipX = FlipX;
            _q.FlipY = FlipY;

            if ((_q.TextureRepeatX == 1.0f) && (_scrollX == 0.0f) && (_q.TextureOffsetX == 0))
                _q.TexAddressU = TextureAddressMode.Clamp;
            else
                _q.TexAddressU = TextureAddressMode.Wrap;

            if ((_q.TextureRepeatY == 1.0f) && (_scrollY == 0.0f) && (_q.TextureOffsetY == 0))
                _q.TexAddressV = TextureAddressMode.Clamp;
            else
                _q.TexAddressV = TextureAddressMode.Wrap;

            // forward to quad
            _q.Render(ObjToWorld, VisibilityLevel, srs);

            Profiler.Instance.EndBlock("T2DScroller.Render");
        }

        public virtual void UpdateAnimation(float dt)
        {
            Profiler.Instance.StartBlock("T2DScroller.UpdateAnimation");

            // Calculate texel shift per world-unit.
            float scrollTexelX = _q.TextureRepeatX / Size.X;
            float scrollTexelY = _q.TextureRepeatY / Size.Y;

            float scrollOffsetX = scrollTexelX * _scrollX * dt;
            float scrollOffsetY = scrollTexelY * _scrollY * dt;

            // Calculate new offset.
            _q.TextureOffsetX = (_q.TextureOffsetX + scrollOffsetX) % 1.0f;
            _q.TextureOffsetY = (_q.TextureOffsetY + scrollOffsetY) % 1.0f;

            Profiler.Instance.EndBlock("T2DScroller.UpdateAnimation");
        }

        public void Dispose()
        {
            _q.Dispose();
        }

        #endregion

        //======================================================
        #region Private, protected, internal fields

        private float _scrollX;
        private float _scrollY;

        private RenderQuad _q = new RenderQuad();

        #endregion
    }
}
